/*
 *	CLIENT routines for Simple Mail Transfer Protocol ala RFC821
 *	A.D. Barksdale Garbee II, aka Bdale, N3EUA
 *	Copyright 1986 Bdale Garbee, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 * 	Modified 14 June 1987 by P. Karn for symbolic target addresses,
 *	also rebuilt locking mechanism
 *	Copyright 1987 1988 David Trulli, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 */

/****************************************************************************
*	$Id: smtpcli.c 1.21 2002/02/01 18:31:10 Simon Exp $
*	29 May 92	1.2		GT	mailroute (): use gateway if defined.			*
*	02 Jun 92	1.3		GT	Add "smtp wait" and timeout on connect () call	*
*							in smtp_send ().								*
*	04 Jun 92	1.4		GT	Make smtptick () a separate process.			*
*	05 Jun 92	1.5		GT	Timeout on connect () to Gateway.				*
*	26 Jun 92			Paul@wolf.demon.co.uk Fixed routing for internal	*
*						email (doesn't go via the external gateway)         *
*	13 Jul 92	1.7		GT	Remove 1.6 changes.								*
*	05 Aug 92	    	amc@beryl.demon.co.uk changed printfs to tprintf	*
*	27 Aug 92   1.10    mt@kram.demon.co.uk added smtp separator            *
*	Dec 92				mt@kram.org don't use gateway for any mail MXed via *
*						us. Add delay channel for local mail  				*
*						Add smtp beep on|off								*
*	19 Apr 93	1.16	GT	If newproc () fails, reset tick flag.			*
*	08 May 93	1.17	GT	Fix warnings.									*
*							Fix "smtp list".								*
*						IAY	Fix dot transparency as per RFC 821.			*
*	07 Nov 93	1.18	GT	Verbose: show "sending" progress.				*
*							mailroute (): cope with domain literals.		*
*	   May 97	1.19	GeK	"smtp list" stat() fix,							*
*	         	    	   	4xx response to RCPT TO: fixed					*
*	         	    	   	include dest IP addr in bounce error message	*
*	         	    	   	"smtp gateway" stores hostname not IP addr		*
*	         	    	   	Limit To: addresses display in smtp list, etc.	*
*	         	    	   	Added "smtp server" cmd for SMTP server			*
*	   Aug 97	1.20	GeK	smtp_gateway global for server %-hack fix;		*
*	         	    	   	Added IP addr (back) to connect fail messages;	*
*	         	    	   	enabled ESMTP; tidied delay channel code; added	*
*	         	    	   	fstat_file() function; log sent ID from server	*
*	01 Feb 02	1.21	SCT	Added YAN-like "$smarthost" subsitution			*
****************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <setjmp.h>
#ifdef UNIX
#include <sys/types.h>
#endif
#ifdef	AMIGA
#include <stat.h>
#else
#include <sys/stat.h>
#endif
#ifdef	__TURBOC__
#include <dir.h>
#include <io.h>
#endif
#include "global.h"
#ifdef	ANSIPROTO
#include <stdarg.h>
#endif
#include "config.h"
#include "domain.h"
#include "mbuf.h"
#include "cmdparse.h"
#include "proc.h"
#include "socket.h"
#include "timer.h"
#include "netuser.h"
#include "smtp.h"
#include "dirutil.h"
#include "commands.h"
#include "session.h"
#include "ip.h"
#include "main.h"	/*  GeK - main_exit prototype  */

#define	USE_ESMTP			/*  GeK - use EHLO as recommended by current drafts  */
#define	LOG_SMTPCLI_ERRORS	/*  GeK - logs 4xx or 5xx responses  */
#define	SMTP_SERVER_CMD		/*  GeK - adds 'smtp server ...' commands  */

#define	MAX_RCPT_DISPLAY	4	/*  GeK - max to: addresses displayed  */

static struct timer Smtpcli_t;
/*  GeK - store gateway name, not address; global for SMTP server %-hack fix  */
char	*smtp_gateway = NULL;
static int32 Delaytime = 0;
int smtpverbose = 1;
char *smtp_separator = NULL;

#ifdef SMTPTRACE
static unsigned short Smtptrace = 0;	/* used for trace level */
static int dosmtptrace __ARGS ((int argc, char *argv[], void *p));

#endif

static unsigned short Smtpmaxcli = MAXSESSIONS;	/* the max client connections
												 * allowed */
static int Smtpsessions = 0; /* number of client connections currently open */
static int Smtpbatch;
int Smtpmode = 0;
int Smtpbeep = 0;

int32 connect_wait_val = 60000L;		/* default connection timeout		 */

static struct smtpcli *cli_session[MAXSESSIONS];		/* queue of client
														 * sessions  */

static void del_job __ARGS ((struct smtp_job * jp));
static void del_session __ARGS ((struct smtpcli * cb));
static int dogateway __ARGS ((int argc, char *argv[], void *p));
static int doseparator __ARGS ((int argc, char *argv[], void *p));
static int dosmtpmaxcli __ARGS ((int argc, char *argv[], void *p));
static int dotimer __ARGS ((int argc, char *argv[], void *p));
static int dodelay __ARGS ((int argc, char *argv[], void *p));
static int dowait __ARGS ((int argc, char *argv[], void *p));
static int dosmtpkill __ARGS ((int argc, char *argv[], void *p));
static int dosmtplist __ARGS ((int argc, char *argv[], void *p));
static int dobatch __ARGS ((int argc, char *argv[], void *p));
static int dobeep __ARGS ((int argc, char *argv[], void *p));
static void execjobs __ARGS ((void));
static int getresp_in_buf __ARGS ((int s, char *buf, int bufsize));		/*  GeK  */
static int getresp __ARGS ((struct smtpcli * ftp, int mincode, char **response));
static void logerr __ARGS ((struct smtpcli * cb, char *line));
static struct smtpcli *lookup __ARGS ((int32 destaddr));
static struct smtpcli *lookup_destname (const char * const destname);	/*  GeK  */
static struct smtpcli *newcb __ARGS ((void));
static int next_job __ARGS ((struct smtpcli * cb));
static int retmail __ARGS ((struct smtpcli * cb));
static void sendcmd __ARGS ((struct smtpcli * cb, char *fmt,...));
static int smtpsendfile __ARGS ((struct smtpcli * cb));
static int setsmtpmode __ARGS ((int argc, char *argv[], void *p));
static struct smtp_job *setupjob __ARGS ((struct smtpcli * cb, char *id, char *from));
static void smtp_send __ARGS ((int unused, void *cb1, void *p));
static int smtpkick __ARGS ((int argc, char *argv[], void *p));
static void smtptick_body __ARGS ((int unused, void *t, void *p));
static int doverbose __ARGS ((int argc, char *argv[], void *p));
static int should_enter_delay __ARGS ((struct smtpcli * cb));
static int explock __ARGS ((char *dir, char *pre, char *suf, int32 age));
static int fstat_file (const char *filepath, struct stat *statp);

#ifdef	USE_ESMTP		/*  GeK  */
static int get_ehlo_resp (struct smtpcli *cb, int mincode);
#endif

static struct cmds Smtpcmds[] = {
	{ "batch",		dobatch,		0,	0,	NULLCHAR },
	{ "beep",		dobeep,			0,	0,	NULLCHAR },
	{ "delay",		dodelay,		0,	0,	NULLCHAR },
	{ "gateway",	dogateway,		0,	0,	NULLCHAR },
	{ "kick",		smtpkick,		0,	0,	NULLCHAR },
	{ "kill",		dosmtpkill,		0,	2,	"kill <jobnumber>" },
	{ "list",		dosmtplist,		0,	0,	NULLCHAR },
	{ "maxclients",	dosmtpmaxcli,	0,	0,	NULLCHAR },
	{ "mode",		setsmtpmode,	0,	0,	NULLCHAR },
	{ "separator",	doseparator,	0,	0,	NULLCHAR },
#ifdef	SMTP_SERVER_CMD		/*  GeK - in smtpserv.c  */
	{ "server",		doserver,		0,	0,	NULLCHAR },
#endif
	{ "timer",		dotimer,		0,	0,	NULLCHAR },
#ifdef SMTPTRACE
	{ "trace",		dosmtptrace,	0,	0,	NULLCHAR },
#endif
	{ "wait",		dowait,			0,	0,	NULLCHAR },
	{ "verbose",	doverbose,		0,	0,	NULLCHAR },		/* Actually for Server */
	{ NULLCHAR },
};

static void stlog(const char *fmt, ...)
{
	va_list ap;
	char buf[256];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	log(-1, "%s", buf);
	tprintf("ST: %s\n", buf);
}

int
    dosmtp (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	return subcmd (Smtpcmds, argc, argv, p);
	}

static int
    dobatch (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	return setbool (&Smtpbatch, "SMTP batching", argc, argv);
	}

static int
    dobeep (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	return setbool (&Smtpbeep, "SMTP beep on delivery", argc, argv);
	}

static int
    dosmtpmaxcli (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	return setshort (&Smtpmaxcli, "Max clients", argc, argv);
	}

static int
    setsmtpmode (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	if (argc < 2)
		{
		tprintf ("smtp mode: %s\n",
				 (Smtpmode & QUEUE) ? "queue" : "route");
		}
	else
		{
		switch (*argv[1])
			{
			case 'q':
				Smtpmode |= QUEUE;
				break;
			case 'r':
				Smtpmode &= ~QUEUE;
				break;
			default:
				tprintf ("Usage: smtp mode [queue | route]\n");
				break;
			}
		}

	return 0;
	}
static int
    dogateway (argc, argv, p)
int argc;
char *argv[];
void *p;

	{

	if (argc < 2)
		{
		/*  GeK - show stored hostname  */
		tprintf ("%s\n", smtp_gateway != NULL ? smtp_gateway : "(none)");
		}
	else
		/*  GeK - only resolve hostname when required  */
		{
		free (smtp_gateway);
		/*  GeK - provide method of undefining gateway  */
		smtp_gateway = ((stricmp (argv[1], "none") == 0) ? NULL : strdup (argv[1]));
		}

	return 0;
	}

static int
    doseparator (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	int i;
	char *t;

	if (argc < 2)
		{
		if (smtp_separator)
			free (smtp_separator);

		smtp_separator = NULL;
		}
	else
		{
		t = mallocw (81);
		if (t == NULL)
			return 1;

		t[0] = '\0';
		for (i = 1; i < argc; i++)
			{
			strcat (t, argv[i]);
			if (i < (argc - 1))
				strcat (t, " ");

			}
		if ((smtp_separator = mallocw (strlen (t) + 1)) == NULL)
			{
			free (t);
			return 1;
			}

		strcpy (smtp_separator, t);
		free (t);
		}

	return 0;
	}

#ifdef SMTPTRACE
static int
    dosmtptrace (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	return setshort (&Smtptrace, "SMTP tracing", argc, argv);
	}

#endif

/* list jobs wating to be sent in the mqueue */

static int dosmtplist (argc, argv, p)
int argc;
char *argv[];
void *p;
	{
	char tstring[80];
	char line[20];
	char host[LINELEN];
	char to[LINELEN];
	char from[LINELEN];
	char *cp;
	char status;
	struct stat stbuf;
	struct tm *tminfo;
	FILE *fp;
	int to_count;	/*  GeK - counts number of recipients to limit display  */

	Current->flowmode = 1;						 /* Enable the more mechanism */
	tprintf ("S     Job    Size Date  Time  Host                 From\n");
	filedir (Mailqueue, 0, line);
	while (line[0] != '\0')
		{
		(void) sprintf (tstring, "%s/%s", Mailqdir, line);
		if ((fp = fopen (tstring, READ_TEXT)) == NULLFILE)
			{
			tprintf ("Can't open %s: %s\n", tstring, sys_errlist[errno]);
			pwait (NULL);
			filedir (Mailqueue, 1, line);
			continue;
			}

		if ((cp = strrchr (line, '.')) != NULLCHAR)
			*cp = '\0';

		(void) sprintf (tstring, "%s/%s.lck", Mailqdir, line);
		if (access (tstring, 0))
			status = ' ';
		else
			status = 'L';

		/* Now see if it's in delay */
		if (status == ' ')
			{
			sprintf (tstring, "%s/%s.dly", Mailqdir, line);
			if (!access (tstring, 0))
				status = 'D';

			}

		(void) sprintf (tstring, "%s/%s.txt", Mailqdir, line);
		/*  GeK - replaced stat() with helper function calling fstat()  */
		if (fstat_file (tstring, &stbuf))
			{
			tprintf ("Can't stat %s: %s\n", tstring, sys_errlist[errno]);
			goto next_file;		/* GeK - ensures wrk file closed  */
			}

		tminfo = localtime (&stbuf.st_ctime);
		fgets (host, sizeof (host), fp);
		rip (host);
		fgets (from, sizeof (from), fp);
		rip (from);
		tprintf ("%c %7s %7ld %02d/%02d %02d:%02d %-20s %s\n   ",
				 status, line, stbuf.st_size,
				 tminfo->tm_mday,		/*  GeK - swapped month/day over  */
				 tminfo->tm_mon + 1,
				 tminfo->tm_hour,
				 tminfo->tm_min,
				 host, from);

		to_count = 0;
		while (fgets (to, sizeof (to), fp) != NULLCHAR)
			{
			/*  GeK - limit number of recipients displayed  */
			if (++to_count > MAX_RCPT_DISPLAY)
				{
				tputs (" ...");
				break;
				}
			rip (to);
			tprintf (" %s", to);
			}

		tputc ('\n');
next_file:
		fclose (fp);
		pwait (NULL);
		filedir (Mailqueue, 1, line);
		}

	Current->flowmode = 0;
	return 0;
	}	/* static int dosmtplist (argc, argv, p) */
	

/* kill a job in the mqueue */
static int
    dosmtpkill (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	char s[SLINELEN];
	char *cp, c;

	sprintf (s, "%s/%s.lck", Mailqdir, argv[1]);
	cp = strrchr (s, '.');
	if (!access (s, 0))
		{
		Current->ttystate.echo = Current->ttystate.edit = 0;
		c = keywait ("Warning, the job is locked by SMTP. Remove (y/n)? ", 0);
		Current->ttystate.echo = Current->ttystate.edit = 1;
		if (c != 'y')
			return 0;

		(void) unlink (s);
		}

	strcpy (cp, ".wrk");
	if (unlink (s))
		tprintf ("Job id %s not found\n", argv[1]);

	strcpy (cp, ".txt");
	(void) unlink (s);
	strcpy (cp, ".dly");
	(void) unlink (s);
	return 0;
	}

/* Set outbound spool scan interval */
static int
    dotimer (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	if (argc < 2)
		{
		tprintf ("%lu/%lu\n",
				 read_timer (&Smtpcli_t) / 1000L,
				 dur_timer (&Smtpcli_t) / 1000L);
		return 0;
		}

	Smtpcli_t.func = (void (*) ()) smtptick;	 /* what to call on timeout */
	Smtpcli_t.arg = NULL;						 /* dummy value */
	set_timer (&Smtpcli_t, atol (argv[1]) * 1000L);		/* set timer duration */
	start_timer (&Smtpcli_t);					 /* and fire it up */
	return 0;
	}

/* Set delay channel time */
static int
    dodelay (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	/*  GeK - use built in function  */
	return setlong (&Delaytime, "SMTP delay", argc, argv);
	}

/****************************************************************************
*	dowait																	*
*	Displays or sets the connection wait timeout.							*
****************************************************************************/

static int dowait (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	if (argc < 2)
		{
		tprintf ("smtp connection timeout: %lu\n", connect_wait_val / 1000L);
		return (0);
		}

	connect_wait_val = atol (argv[1]) * 1000L;	 /* set timeout					 */
	return (0);
	}						 /* static int dowait (argc, argv, p) */

static int
    smtpkick (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	char	*cp = NULL;

	if (argc > 1)
		cp = strdup (argv[1]);

	smtptick ((void *) cp);
	return 0;
	}

/* This is the routine that gets called every so often to do outgoing
 * mail processing. When called with a null argument, it runs the entire
 * queue; if called with a specific non-zero IP address from the remote
 * kick server, it only starts up sessions to that address.
 */

static char smtptick_running = 0;

int
    smtptick (t)
void *t;
	{
	if (smtptick_running != 0)
	{
		stlog("smtp_tick process already running");
		return (0);
	}

	smtptick_running = 1;
	stlog("Starting smtp_tick process");
	if (newproc ("smtp_tick", 1024, smtptick_body, 0, t, NULL, 0) == NULLPROC)
		{
		/* Wait a while. */

		start_timer (&Smtpcli_t);
		smtptick_running = 0;
		free (t);
		}

	return (0);
	}

static void smtptick_body (unused, t, p)
int unused;
void *t;
void *p;
	{
	register struct smtpcli *cb;
	struct smtp_job *jp;
	struct list *ap;
	char *buf, wfilename[13], prefix[9];
	char *cp, *cp1;
	int32 destaddr, num_files;
	char *target;	/*  GeK - target now a string ptr  */
	FILE *wfile;
	int to_count;	/*  GeK - used to count recipients  */

	stlog("Entered smtptick_body()");
	target = (char *) t;
#ifdef SMTPTRACE
	if (Smtptrace > 5)
		tprintf ("smtp daemon entered, target = %s\n", target != NULL ? target : "(none)");
#endif

	if (availmem () < Memthresh)
		{
		/* Memory is tight, don't do anything */
		/* Restart timer */
		start_timer (&Smtpcli_t);
		smtptick_running = 0;
		free (target);
		stlog("Leaving smtptick_body(): not enough memory");
		return;
		}

	buf = mallocw (LINELEN);

	/* First see how many there are now */

	num_files = 0;
	for (filedir (Mailqueue, 0, wfilename); wfilename[0] != '\0';
		 filedir (Mailqueue, 1, wfilename))
		num_files++;

	stlog("Found %d queued messages to send", num_files);
	/* Try to deal with that many messages, no more or we may loop forever */
	for (filedir (Mailqueue, 0, wfilename); wfilename[0] != '\0' && num_files--;
		 filedir (Mailqueue, 1, wfilename))
		{

		/* save the prefix of the file name which it job id */
		cp = wfilename;
		cp1 = prefix;
		while (*cp && *cp != '.')
			*cp1++ = *cp++;

		*cp1 = '\0';

		/* lock this file from the smtp daemon */
		if (mlock (Mailqdir, prefix))
			continue;

		/* Try to bring the job outof delay (in case it's in delay) */
		/* GeK - also check that it's not in delay */
		if (explock (Mailqdir, prefix, "dly", Delaytime) < 0)
			{
			/*  Delay lock exists but not deleted  */
			rmlock (Mailqdir, prefix);
			continue;
			}

		/*  GeK - no longer need to check delay lock since included above  */

		sprintf (buf, "%s/%s", Mailqdir, wfilename);
		if ((wfile = fopen (buf, READ_TEXT)) == NULLFILE)
			{
			/* probably too many open files */
			(void) rmlock (Mailqdir, prefix);
			/* continue to next message. The failure may be temporary */
			continue;
			}

		stlog("Msg %s: WRK file opened OK", prefix);
		(void) fgets (buf, LINELEN, wfile);/* read target host */
		rip (buf);

		stlog("Msg %s: WRK target host \"%s\"", prefix, buf);
		tprintf("About to do SMARTHOST code\n");
#ifdef	SMARTHOST		/* Target "$smarthost" -> SMTP gateway */
		if (smtp_gateway != NULL && stricmp (buf, "$smarthost") == 0)
		{
			stlog("SMARTHOST: replacing $smarthost by gateway \"%s\"", smtp_gateway);
			strcpy (buf, smtp_gateway);		/* should be OK: buf is 256 long */
		}
		else if (smtp_gateway != NULL)
			stlog("SMARTHOST: gateway defined, but target not $smarthost");
		else
			stlog("SMARTHOST: no SMTP gateway; SMARTHOST mode disabled");
#endif
		tprintf("Done SMARTHOST code\n");
		stlog("Msg %s: final target host \"%s\"", prefix, buf);

		/*  GeK - look for other destnames first  */
		cb = lookup_destname (buf);
		if (cb == NULL)
			{
			if (target != NULL && stricmp (target, buf) != 0)
				{
				fclose (wfile);
				(void) rmlock (Mailqdir, prefix);
				continue;							 /* Not the proper target of
													  * a kick */
				}

			if ((destaddr = mailroute (buf, Ip_addr)) == 0)
				{
				tprintf ("** smtp: Unknown address %s\n", buf);
				fclose (wfile);
				(void) rmlock (Mailqdir, prefix);
				continue;
				}

			stlog("Msg %s: mailroute() returned address %s", prefix, inet_ntoa(destaddr));
			cb = lookup (destaddr);
			}

		if (cb == NULL)
			{
			/* there are enough processes running already */
			if (Smtpsessions >= Smtpmaxcli)
				{
#ifdef SMTPTRACE
				if (Smtptrace)
					{
					tprintf ("smtp daemon: too many processes\n");
					}
#endif

				fclose (wfile);
				(void) rmlock (Mailqdir, prefix);
				break;
				}

			if ((cb = newcb ()) == NULLSMTPCLI)
				{
				fclose (wfile);
				(void) rmlock (Mailqdir, prefix);
				break;
				}

			cb->ipdest = destaddr;
			cb->destname = strdup (buf);
			}
		else
			{
			if (cb->lock)
				{
				/* This system is already is sending mail lets not
				 * interfere with its send queue. */
				fclose (wfile);
				(void) rmlock (Mailqdir, prefix);
				continue;
				}
			}

		stlog("Msg %s: cb \"%s\" [%s]", prefix, cb->destname, inet_ntoa(cb->ipdest));
		(void) fgets (buf, LINELEN, wfile);	 /* read from */
		rip (buf);
		/*  GeK - also quit here if not enough memory  */
		if (availmem() < (4 * Memthresh)
		    || (jp = setupjob (cb, prefix, buf)) == NULLJOB)
			{
			del_session (cb);
			fclose (wfile);
			(void) rmlock (Mailqdir, prefix);
			break;
			}

		while (fgets (buf, LINELEN, wfile) != NULLCHAR)
			{
			rip (buf);

			/*  GeK - skip empty forward path  */
			if (*buf != '\0')
				addlist (&jp->to, buf, DOMAIN);		/*  GeK - addlist can't fail  */
			}
		fclose (wfile);

		if (smtpverbose != 0)
			{
			to_count = 0;	/*  GeK - counts number of recipients  */

			/* Show sending progress. */

			tprintf ("Sending %s From: %s To:", prefix, jp->from);

			for (ap = jp->to; ap != NULL; ap = ap->next)
				if (++to_count > MAX_RCPT_DISPLAY)	/*  GeK - limit addrs displayed  */
					{
					tputs (" ...");
					break;	/*  for ()  */
					}
				else
					tprintf (" %s", ap->val);

			tputc ('\n');
			}	/* if (smtpverbose != 0 && Smtptrace <= 1) */

		}

	/* start sending that mail */
	stlog("Calling execjobs()");
	execjobs ();

	/* Restart timer */
	start_timer (&Smtpcli_t);
	smtptick_running = 0;
	free (buf);
	free (target);
	stlog("Leaving smtptick_body()");
	return;
	}

static int
    should_enter_delay (cb)
struct smtpcli *cb;
	{
#ifdef	notdef	/*  GeK  */
	resolve_mx (cb->destname);
	if (are_we_an_mx)
		return 1;
#endif

	/*  GeK - this is now a function of its own in domain.c
	 *  It should speed up both this lookup and resolve_mx() lookups
	 *  since they require differing lookups for the MX list
	 */
	return (are_we_an_mx (cb->destname));
	}

/*  GeK - adjusted to take a 8 char prefix rather than value of msgid  */
void
delay_job (char *prefix)
{
	if (Delaytime == 0)
		/* Not using SMTP delay */
		return;

#ifdef SMTPTRACE
	if (Smtptrace)
		tprintf ("Delivery Failed: adding job %s to delay queue\n", prefix);
#endif

	mlock_suffix (Mailqdir, prefix, "dly");
}

/* This is the master state machine that handles a single SMTP transaction.
 * It is called with a queue of jobs for a particular host.
 * The logic is complicated by the "Smtpbatch" variable, which controls
 * the batching of SMTP commands. If Smtpbatch is true, then many of the
 * SMTP commands are sent in one swell foop before waiting for any of
 * the responses. Unfortunately, this breaks many brain-damaged SMTP servers
 * out there, so provisions have to be made to operate SMTP in lock-step mode.
 */
static void
smtp_send (int unused, void *cb1, void *p)
{
	struct smtpcli *cb;
	struct list *tp, *tp_next, *tp_last;
	struct list *retry_list = NULL;
	struct list *fail_list = NULL;
	struct sockaddr_in fsocket;
	char *cp;
	int rcode;
	int goodrcpt;
//	int i;
#ifndef	USE_ESMTP	/*  init isn't needed with ESMTP defined  */
	int init = 1;
#endif
	char *response = NULL;			/*  GeK - holds server response after sending message  */
	int	  using_gateway = FALSE;	/*  GeK - TRUE when talking to gateway  */
	long  response_timeout = (long) 5 * connect_wait_val;	/*  nominal 5 minutes  */
	int	  batch = 0 /* Smtpbatch */;		/*  GeK - snapshot copy of batch flag  */

	cb = (struct smtpcli *) cb1;
	cb->lock = 1;
	fsocket.sin_family = AF_INET;
	fsocket.sin_addr.s_addr = cb->ipdest;
	fsocket.sin_port = IPPORT_SMTP;

	cb->s = socket (AF_INET, SOCK_STREAM, 0);
	/*  GeK - socket can fail  */
	if (cb->s < 0)
		{
		cb->lock = 0;
		return;
		}
	sockmode (cb->s, SOCK_ASCII);
	setflush (cb->s, -1);						 /* We'll explicitly flush
												  * before reading */
#ifdef SMTPTRACE
	if (Smtptrace)
		tprintf ("SMTP client trying %s ...\n", inet_ntoa (cb->ipdest));	/*  GeK - add IP addr  */
#endif

	/* Run connect () under a timeout in case the destination is
	 * unreachable. */

	alarm (connect_wait_val);
	if (connect (cb->s, (char *) &fsocket, SOCKSIZE) == 0)
		{
		alarm (0L);
#ifdef SMTPTRACE
		if (Smtptrace)
			tprintf ("SMTP Connected to %s\n", inet_ntoa (cb->ipdest));	/*  GeK - add IP addr  */
#endif
		}
	else
		{
		/*  GeK - failure logging moved from below  */
		cp = (!run_timer (&Curproc->alarm)) ? "timeout" : sockerr (cb->s);
		if (cp == NULL)
			cp = "";
		alarm (0L);
#ifdef SMTPTRACE
		if (Smtptrace)
			tprintf ("SMTP %s Connect failed: %s%c", psocket (&fsocket), cp, '\n');
#endif
		/*  GeK - put psocket() back  */
		log (cb->s, "SMTP %s Connect failed: %s%c", psocket (&fsocket), cp, '\0');

try_gateway:
		if (Delaytime && (cb->ipdest == Ip_addr || should_enter_delay (cb)))
			{
			/*  GeK - want to delay all jobs in the queue, not just the first one  */
			struct smtp_job *jp;

			jp = cb->jobq;
			while (jp != NULL)
				{
				delay_job (jp->jobname);	/*  GeK - pass prefix, not value  */
				jp = jp->next;
				}

			/*  GeK - this went into the banner check on a unconnected socket .. hmmm  */
			goto quit_no_cmd;
			}
		else
		/*  GeK - only resolve gateway when reqd  */
		if (smtp_gateway != NULL && (cb->ipdest = resolve (smtp_gateway)) != 0)
			{
			/* Try sending it via the gateway, as long as it's not for us. */
#ifdef SMTPTRACE
			if (Smtptrace)
				/*  GeK - show gateway details  */
				tprintf ("SMTP client trying gateway %s [%s] ...\n", smtp_gateway, inet_ntoa (cb->ipdest));
#endif

			(void) close_s (cb->s);
			/*  GeK - Gateway not used, ipdest is set above  */
			fsocket.sin_addr.s_addr = cb->ipdest;
			cb->s = socket (AF_INET, SOCK_STREAM, 0);
			sockmode (cb->s, SOCK_ASCII);
			setflush (cb->s, -1);
			alarm (connect_wait_val);
			if (connect (cb->s, (char *) &fsocket, SOCKSIZE) != 0)
				{
				/*  GeK - check for timeout  */
				cp = (!run_timer (&Curproc->alarm)) ? "timeout" : sockerr (cb->s);
				if (cp == NULL)
					cp = "";
				alarm (0L);
#ifdef SMTPTRACE
				if (Smtptrace)
					tprintf ("SMTP %s Connect failed: %s%c", psocket (&fsocket), cp, '\n');
#endif
				log (cb->s, "SMTP %s Connect failed: %s%c", psocket (&fsocket), cp, '\0');

				goto quit_no_cmd;
				}
			else
				{
				alarm (0L);
#ifdef SMTPTRACE
				if (Smtptrace)
					tprintf ("SMTP Connected to gateway %s [%s]\n", smtp_gateway, inet_ntoa (cb->ipdest));
#endif
				using_gateway = TRUE;
				}

			}									 /* if (Gateway != 0L) */
		else	/*  Not delayable, gateway undefined or unresolvable  */
			{
			/*  GeK - failure logging moved above  */
			goto quit_no_cmd;
			}

		}

#ifndef	USE_ESMTP		/*  GeK - don't stream initial response with ESMTP  */
	if (!batch)
#endif
		{
		alarm (response_timeout);	/*  GeK - nominal 5 minute timeout on initial banner  */
		rcode = getresp (cb, 200, NULL);
		alarm (0L);
		/*  GeK - on initial banner failure, fallback to the gateway
		 *        (unless already talking to the gateway)
		 */
		if (rcode < 0 || rcode >= 400)
			{
			if (smtp_gateway == NULL || using_gateway)
				goto quit;
			
			sendcmd (cb, "QUIT\n");
			goto try_gateway;
			}
		}

#ifdef	USE_ESMTP	/*  GeK - attempt ESMTP EHLO first  */

	sendcmd (cb, "EHLO %s\n", Hostname);
	alarm (response_timeout);
	rcode = get_ehlo_resp (cb, 200);
	alarm (0);
	if (rcode < 0 || (rcode >= 400 && rcode < 500))
		/*  timeout/transient failure  */
		goto quit;
	else if (rcode >= 500)
		{
		/*  EHLO unknown, recommended action is a RSET, ignore response, then HELO  */
		sendcmd (cb, "RSET\n");
		alarm (response_timeout);
		rcode = get_ehlo_resp (cb, 200);
		alarm (0);
		if (rcode < 0)
			goto quit;

		sendcmd (cb, "HELO %s\n", Hostname);
		alarm (response_timeout);
		rcode = getresp (cb, 200, NULL);
		alarm (0);
		if (rcode < 0 || rcode >= 400)
			goto quit;
		}

#else	/* !USE_ESMTP */

	/* Say HELO */

	sendcmd (cb, "HELO %s\n", Hostname);
	if (!batch)
		{
		rcode = getresp (cb, 200, NULL);
		/*  GeK - should do something here for 5xx responses to HELO ?  */
		if (rcode == -1 || rcode >= 400)
			goto quit;
		}

#endif	/* !USE_ESMTP */

	do
		{										 /* For each message... */

		/* if this file open fails, skip it */
		if ((cb->tfile = fopen (cb->tname, READ_TEXT)) == NULLFILE)
			continue;

		goodrcpt = 0;		/*  GeK - ensures 0 for MAIL FROM fail 5xx below  */

		free (response);
		response = NULL;	/*  GeK - ensure stored response is empty  */

		/* Send MAIL and RCPT commands */

		sendcmd (cb, "MAIL FROM:<%s>\n", cb->jobq->from);
		if (!batch)
			{
			alarm (response_timeout);
			rcode = getresp (cb, 200, NULL);
			alarm (0);
			if (rcode == -1)
				goto quit;
			if (rcode >= 400)
				goto fail_job;		/*  MAIL FROM failure  */
			}

		tp_last = NULL;
		for (tp = cb->jobq->to; tp != NULL; tp = tp_next)
			{
			tp_next = tp->next;
			sendcmd (cb, "RCPT TO:<%s>\n", tp->val);
			if (!batch)
				{
				alarm (response_timeout);
				rcode = getresp (cb, 200, NULL);
				alarm (0);
				if (rcode == -1)
					goto quit;
				else if (rcode >= 400 && rcode < 500)
					{
					/*  Move from to: list to retry queue  */
					if (tp_last == NULL)
						cb->jobq->to = tp_next;
					else
						tp_last->next = tp_next;
					tp->next = retry_list;
					retry_list = tp;

					/*
					 *	GeK - this is a temporary fudge to avoid not sending
					 *	mail to recipients which result in 400 responses
					 *	- jump out of session.
					 */
					goto quit;
					}
				else if (rcode >= 500)
					{
					/*  Move from to: list to fail queue  */
					if (tp_last == NULL)
						cb->jobq->to = tp_next;
					else
						tp_last->next = tp_next;
					tp->next = fail_list;
					fail_list = tp;
					}
				else if (rcode < 400)
					{
					goodrcpt = 1;				 /* At least one good */
					tp_last = tp;	/*  Last valid recipient  */
					}
				}
			}

		if (!batch && goodrcpt == 0)
			{
			if (retry_list != NULL)
				rcode = 400;	/*  Some worth retrying  */
			else
				rcode = 500;	/*  All failures  */

			goto fail_job;
			}

		/* Send DATA command */
		sendcmd (cb, "DATA\n");

		if (!batch)
			{
			rcode = getresp (cb, 200, NULL);
			if (rcode == -1)
				goto quit;
			if (rcode >= 400)
				goto fail_job;
			}

#ifdef	notdef	/*  GeK - temp removed  */
		if (batch)
			{
			int	fail = FALSE;
				
			/* Now wait for the responses to come back. The first time we
			 * do this, we wait first for the start banner and HELO
			 * response. In any case, we wait for the response to the MAIL
			 * command here. */
#ifndef	USE_ESMTP		/*  GeK - ESMTP doesn't stream initial nor EHLO response  */
			for (i = (init ? 3 : 1); i > 0; i--)
#endif
				{
				rcode = getresp (cb, 200, NULL);
				if (rcode == -1)
					goto quit;
				else if (rcode >= 400 && rcode < 500)
					goto fail_job;	/*  Assumes 4xx quits session below else the state machine is screwed  */
				else if (rcode >= 500)
					fail = TRUE;
				}
#ifndef	USE_ESMTP
			init = 0;
#endif

			/* Now process the responses to the RCPT commands */
			for (tp = cb->jobq->to; tp != NULL; tp = tp_next)
				{
				tp_next = tp->next;
				alarm (response_timeout);
				alarm (0);
				rcode = getresp (cb, 200, NULL);
				if (rcode == -1)
					{
					cb->jobq->to = tp;		/*  Ensure to list pts to remaining valid rcpts  */
					goto quit;
					}
				else if (rcode >= 400 && rcode < 500)
					{
					tp->next = retry_list;
					retry_list = tp;
					}
				else if (rcode >= 500)
					{
					tp->next = fail_list;
					fail_list = tp;
					}
				else if (rcode < 400)
					{
					goodrcpt = 1;				 /* At least one good */
					free (tp->val);
					free (tp);
					}
				}

			/*  To: list is now empty  */
			cb->jobq->to = NULL;

			/* And finally get the response to the DATA command. Some
			 * servers will return failure here if no recipients are valid,
			 * some won't. */
			rcode = getresp (cb, 200, NULL);
			if (rcode == -1 || rcode >= 400)
				goto quit;

			/* check for no good rcpt on the list */

			if (goodrcpt == 0)
				{
				sendcmd (cb, ".\n");			 /* Get out of data mode */
				goto quit;
				}

			if (fail)
				goodrcpt = 0;	/*  Ensures no DATA sent below and msg returned  */
			}
#endif	/* notdef */

		/* Send the file. This also closes it */
		smtpsendfile (cb);

		/* Wait for the OK response, logging mail ID */
		alarm (min ((long) 2 * response_timeout, 600000L));
		rcode = getresp (cb, 200, &response);
		alarm (0);
		if (rcode == -1)
			goto quit;

fail_job:
		if (cb->tfile != NULL)
			{
			fclose (cb->tfile);		/*  GeK - ensure file closed before delete below  */
			cb->tfile = NULL;
			}

		if (rcode >= 400)
			{
			int	rset_rcode;

			/*  Failure of some sort, send RSET  */
			sendcmd (cb, "RSET\n");
			rset_rcode = getresp (cb, 200, NULL);
			if (rset_rcode == -1)
				goto quit;
			if (rset_rcode > rcode)
				rcode = rset_rcode;
			}

		if ((rcode >= 200 && rcode < 300) || rcode >= 500)
			{
			int	fail = FALSE;

			/* if a good transfer or permanent failure remove job */
			if (cb->errlog != NULLLIST)
				fail = retmail (cb);

			/*  GeK - if fail is TRUE here then retmail() was unable to
			 *  to return a bounce to our own Postmaster so assume it's
			 *  a local system problem and don't delete the job.
			 */

			if (!fail)
				{
				unlink (cb->tname);			 /* Unlink the textfile */
				unlink (cb->wname);			 /* unlink workfile */
				}

			if (cb->jobq->to != NULL)	/*  GeK - temp  */
			log (cb->s, "SMTP sent job %s To: %s From: %s (%s)",
				 cb->jobq->jobname, cb->jobq->to->val, cb->jobq->from,
				 (response != NULL && strlen (response) > 4) ? response+4 : "");

			if (smtpverbose != 0 && Smtptrace <= 1)
				{
				/* Show sending progress. */

#if	0
				tprintf ("Sent %s From: %s To: %s\n", cb->jobq->jobname,
						 cb->jobq->from, cb->jobq->to->val);
#endif
				tprintf ("Sent %s\n", cb->jobq->jobname);
				}	/* if (smtpverbose != 0 && Smtptrace <= 1) */

			}	/* if ((rcode >= 200 && rcode < 300) || rcode >= 500) */
		else
			{
			/*  GeK - warn of 4xx job temporary failure  */
			log (cb->s, "SMTP job requeued %s To: %s From: %s",
			     cb->jobq->jobname, (cb->jobq->to != NULL) ? cb->jobq->to->val : "", cb->jobq->from);
			if (smtpverbose)
				tprintf ("SMTP requeued failed job %s\n", cb->jobq->jobname);

			/*  GeK - assume a 4xx reply is fatal to this session  */
			break;	/*  do .. while (next_job)  */
			}

		/*  GeK - clean up recipient lists, next_job cleans jobq->to  */
		del_list (retry_list);
		retry_list = NULL;
		del_list (fail_list);
		fail_list = NULL;

		} while (next_job (cb));

quit:
	sendcmd (cb, "QUIT\n");
quit_no_cmd:
	/*
	 *  GeK - we really don't want to do this here under most circumstances
	 */
#ifdef	notdef
	if (cb->errlog != NULLLIST)
		{
		retmail (cb);
		(void) unlink (cb->wname);				 /* unlink workfile */
		(void) unlink (cb->tname);				 /* unlink text */
		}
#endif

	(void) close_s (cb->s);
	if (cb->tfile != NULLFILE)
		fclose (cb->tfile);

	free (response);		/*  GeK - free up stored response  */
	del_list (retry_list);
	del_list (fail_list);

	cb->lock = 0;
	del_session (cb);
	}

/* create mail lockfile */
int
    mlock (dir, id)
char *dir, *id;
	{
	return mlock_suffix (dir, id, "lck");
	}

int
    mlock_suffix (dir, id, suffix)
char *dir, *id, *suffix;
	{
	char lockname[LINELEN];
	int fd;

#ifdef	MSDOS
	if (strlen (id) > 8)
		{										 /* truncate long filenames */
		id[8] = '\0';
		if (id[7] == '/')
			id[7] = '\0';

		}
#endif
	/* Try to create the lock file in an atomic operation */

	sprintf (lockname, "%s/%s.%s", dir, id, suffix);
#ifdef        AMIGA
	/* don't ask, really, just don't ask... I'd do file locking on an Amiga
	 * much more differently than this. */
	if (access (lockname, 0) == 0)
		return -1;
#endif

	if ((fd = open (lockname, O_WRONLY | O_EXCL | O_CREAT, 0600)) == -1)
		return -1;

	close (fd);
	return 0;
	}

/* remove mail lockfile */
int
    rmlock (dir, id)
char *dir, *id;
	{
	return rmlock_suffix (dir, id, "lck");
	}

int
    rmlock_suffix (dir, id, suffix)
char *dir, *id, *suffix;
	{
	char lockname[LINELEN];

#ifdef	MSDOS
	if (strlen (id) > 8)
		{										 /* truncate long filenames */
		id[8] = '\0';
		if (id[7] == '/')
			id[7] = '\0';

		}
#endif

	sprintf (lockname, "%s/%s.%s", dir, id, suffix);
	return (unlink (lockname));
	}

/* free the message struct and data */
static void
     del_session (cb)
register struct smtpcli *cb;
	{
	register struct smtp_job *jp, *tp;
	register int i;

	if (cb == NULLSMTPCLI)
		return;

	for (i = 0; i < MAXSESSIONS; i++)
		if (cli_session[i] == cb)
			{
			cli_session[i] = NULLSMTPCLI;
			break;
			}

	free (cb->wname);
	free (cb->tname);
	free (cb->destname);
	for (jp = cb->jobq; jp != NULLJOB; jp = tp)
		{
		tp = jp->next;
		del_job (jp);
		}

	del_list (cb->errlog);
	free ((char *) cb);
	Smtpsessions--;								 /* number of connections
												  * active */
	}

static void
     del_job (jp)
register struct smtp_job *jp;
	{
	if (*jp->jobname != '\0')
		(void) rmlock (Mailqdir, jp->jobname);

	free (jp->from);
	del_list (jp->to);
	free ((char *) jp);
	}

/* delete a list of list structs */
void
     del_list (lp)
struct list *lp;
	{
	register struct list *tp, *tp1;

	for (tp = lp; tp != NULLLIST; tp = tp1)
		{
		tp1 = tp->next;
		free (tp->val);
		free ((char *) tp);
		}
	}

/* stub for calling mdaemon to return message to sender */
static int
     retmail (cb)
struct smtpcli *cb;
	{
	FILE *infile;
	int fail = FALSE;
	static char msg_success[] = "SMTP returned job %s to %s%c";
	static char msg_fail[] = "SMTP failed to return job %s to %s%c";

	/*  GeK - don't do anything for a null return path  */
	if (*cb->jobq->from == '\0')
		return (FALSE);

#ifdef SMTPTRACE
	if (Smtptrace > 5)
		{
		tprintf ("smtp job %s returned to sender\n", cb->jobq->jobname);
		}
#endif

	if ((infile = fopen (cb->tname, READ_TEXT)) == NULLFILE)
		return (TRUE);

	if (mdaemon (infile, cb->jobq->from, cb->errlog, 1))
		{
		/*  GeK - mdaemon failed; as a last ditch attempt, mail it to Postmaster with explanation  */
		static const char msg[] = "Mailer daemon was unable to return the following error message to\n<%s>:\n";
		char             *buf;
		struct list      *lp;

		lp = (struct list *) callocw (1, sizeof (struct list));
		lp->val = mallocw (sizeof (msg) + strlen (cb->jobq->from));
		sprintf (lp->val, msg, cb->jobq->from);
		/*  Place at head of error log  */
		lp->next = cb->errlog;
		cb->errlog = lp;

		buf = mallocw (strlen (Hostname) + 12);
		sprintf (buf, "Postmaster@%s", Hostname);
		fail = mdaemon (infile, buf, cb->errlog, 1);
		log (cb->s, fail ? msg_fail : msg_success, cb->jobq->jobname, fail ? cb->jobq->from : buf, '\0');
		if (smtpverbose)
			tprintf (fail ? msg_fail : msg_success, cb->jobq->jobname, fail ? cb->jobq->from : buf, '\n');
		free (buf);
		}
	else
		{
		log (cb->s, msg_success, cb->jobq->jobname, cb->jobq->from, '\0');
		if (smtpverbose)
			tprintf (msg_success, cb->jobq->jobname, cb->jobq->from, '\n');
		}
	fclose (infile);
	return (fail);
	}

/* look to see if a smtp control block exists for this ipdest */
static struct smtpcli *
        lookup (destaddr)
int32 destaddr;
	{
	register int i;

	for (i = 0; i < MAXSESSIONS; i++)
		{
		if (cli_session[i] == NULLSMTPCLI)
			continue;

		if (cli_session[i]->ipdest == destaddr)
			return cli_session[i];

		}
	return NULLSMTPCLI;
	}

/*  GeK - version of lookup() to find SMTP control block for dest address  */
static struct smtpcli *
lookup_destname (const char * const destname)
{
	int	i;

	for (i = 0; i < MAXSESSIONS; i++)
		{
		if (cli_session[i] != NULL
		    && cli_session[i]->destname != NULL
		    && stricmp (cli_session[i]->destname, destname) == 0)
			return (cli_session[i]);
		}

	return (NULL);
}

/* create a new  smtp control block */
static struct smtpcli *
        newcb ()
	{
	register int i;
	register struct smtpcli *cb;

	for (i = 0; i < MAXSESSIONS; i++)
		{
		if (cli_session[i] == NULLSMTPCLI)
			{
			cb = (struct smtpcli *) callocw (1, sizeof (struct smtpcli));
			/*  GeK - added +1 to account for directory slash  */
			cb->wname = mallocw ((unsigned) strlen (Mailqdir) + JOBNAME + 1);
			cb->tname = mallocw ((unsigned) strlen (Mailqdir) + JOBNAME + 1);
			cli_session[i] = cb;
			Smtpsessions++;						 /* number of connections
												  * active */
			return (cb);
			}

		}
	return NULLSMTPCLI;
	}

static void
     execjobs ()
	{
	register struct smtpcli *cb;
	register int i;

	for (i = 0; i < MAXSESSIONS; i++)
		{
		cb = cli_session[i];
		if (cb == NULLSMTPCLI)
			continue;

		if (cb->lock)
			continue;

		sprintf (cb->tname, "%s/%s.txt", Mailqdir, cb->jobq->jobname);
		sprintf (cb->wname, "%s/%s.wrk", Mailqdir, cb->jobq->jobname);

		newproc ("smtp_send", 1024, smtp_send, 0, cb, NULL, 0);

#ifdef SMTPTRACE
		if (Smtptrace)
			tprintf ("Trying Connection to %s\n", inet_ntoa (cb->ipdest));
#endif

		}
	}

/* add this job to control block queue */
static struct smtp_job *
         setupjob (cb, id, from)
struct smtpcli *cb;
char *id, *from;
	{
	register struct smtp_job *p1, *p2;

	p1 = (struct smtp_job *) callocw (1, sizeof (struct smtp_job));
	p1->from = strdup (from);
	strcpy (p1->jobname, id);
	/* now add to end of jobq */
	if ((p2 = cb->jobq) == NULLJOB)
		cb->jobq = p1;
	else
		{
		while (p2->next != NULLJOB)
			p2 = p2->next;

		p2->next = p1;
		}

	return p1;
	}

/* called to advance to the next job */
static int
    next_job (cb)
register struct smtpcli *cb;
	{
	register struct smtp_job *jp;

	jp = cb->jobq->next;
	del_job (cb->jobq);
	/* remove the error log of previous message */
	del_list (cb->errlog);
	cb->errlog = NULLLIST;
	cb->jobq = jp;
	if (jp == NULLJOB)
		return 0;

	sprintf (cb->tname, "%s/%s.txt", Mailqdir, jp->jobname);
	sprintf (cb->wname, "%s/%s.wrk", Mailqdir, jp->jobname);
#ifdef SMTPTRACE
	if (Smtptrace > 5)
		{
		tprintf ("sending job %s\n", jp->jobname);
		}
#endif

	return 1;

	}

/* Mail routing function. */
int32
mailroute (char *dest,
           int32 gateway)
{
	int32 destaddr;

	/* Test for domain literal first. */
	if (*dest == '[')
		return (aton (dest));

	/* If the destination is this site, then *don't* go through a mail
	 * server (MX site)
	 */
	if (stricmp (dest, Hostname) == 0)
		{
		stlog("mailroute: dest \"%s\" is local hostname", dest);
		/*  GeK - seems odd not to use our IP addr here if we know it  */
		if (Ip_addr != 0)
			return (Ip_addr);
		else if ((destaddr = resolve (dest)) != 0)
			return destaddr;
		}
		
	/*  GeK - if destination matches gateway then bypass MX lookup
	 *  (primarily for speed, it saves a DNS query - if only an A
	 *  record exists for the gateway then resolve_mx() would handle
	 *  it correctly by RFC974 (STD14)).
	 */
	if (smtp_gateway != NULL
	    && stricmp (dest, smtp_gateway) == 0
	    && (destaddr = resolve (smtp_gateway)) != 0)
	{
		stlog("mailroute: dest \"%s\" is SMTP gateway", dest);
		return destaddr;
	}

	/* look up address or use the gateway */

	destaddr = resolve_mx (dest);
	if (destaddr == 0)
		{
		stlog("mailroute: dest \"%s\" MX resolve failed", dest);
#ifdef DSERVER
		if (zone_filename (dest, NULL))
			destaddr = gateway;
		else
#endif
		/*  GeK - SMTP gateway is now resolved when required  */
		if (smtp_gateway != NULL)
			{
			destaddr = resolve (smtp_gateway);	/*  Attempt to use the gateway  */
#ifdef	SMTPTRACE	/*  GeK - added to track reset connections  */
			if (Smtptrace && destaddr != 0)
				tprintf ("mailroute %s failed; using gateway %s\n", dest, smtp_gateway);
#endif	/* SMTPTRACE */
			}
		}
	else
		stlog("mailroute: dest \"%s\" MX resolved OK", dest);

	return destaddr;
}

/* save line in error list */
static void
     logerr (cb, line)
struct smtpcli *cb;
char *line;
	{
	register struct list *lp, *tp;

	tp = (struct list *) callocw (1, sizeof (struct list));
	tp->val = strdup (line);
	/* find end of list */
	if ((lp = cb->errlog) == NULLLIST)
		cb->errlog = tp;
	else
		{
		while (lp->next != NULLLIST)
			lp = lp->next;

		lp->next = tp;
		}
	}

static int
    smtpsendfile (cb)
register struct smtpcli *cb;
	{
	int error = 0;
	int last_had_newline = TRUE;	/*  GeK - prevent breaking long lines  */

	while (fgets (cb->buf, sizeof (cb->buf), cb->tfile) != NULLCHAR)
		{
		/*
		** Dot transparency: add another '.' at the start of
		** any line which already starts with a '.'
		*/

		/*  GeK - only check at the start of new lines  */
		if (last_had_newline && cb->buf[0] == '.')
			usputc (cb->s,'.');

		usputs (cb->s, cb->buf);
		last_had_newline = (strchr (cb->buf, '\n') != NULL);
		}

	fclose (cb->tfile);
	cb->tfile = NULLFILE;
	/* Send the end-of-message command */
	/*  GeK - now know here whether last buf had \n or not  */
	if (last_had_newline)
		sendcmd (cb, ".\n");
	else
		sendcmd (cb, "\n.\n");

	return error;
	}

/* do a printf() on the socket with optional local tracing */
#ifdef	ANSIPROTO
static void
     sendcmd (struct smtpcli * cb, char *fmt,...)
	{
	va_list args;

	va_start (args, fmt);
	vsprintf (cb->buf, fmt, args);		/*  GeK - moved from below, prevents duplication  */

#ifdef	SMTPTRACE
	if (Smtptrace)
		tprintf ("smtp sent: %s", cb->buf);
#endif

	usputs (cb->s, cb->buf);
	va_end (args);
	}

#else
static void
     sendcmd (cb, fmt, arg1, arg2, arg3, arg4)
struct smtpcli *cb;
char *fmt;
int arg1, arg2, arg3, arg4;
	{
#ifdef	SMTPTRACE
	if (Smtptrace)
		{
		tprintf ("smtp sent: ");
		tprintf (fmt, arg1, arg2, arg3, arg4);
		}
#endif

	sprintf (cb->buf, fmt, arg1, arg2, arg3, arg4);
	usputs (cb->s, cb->buf);
	}

#endif

static int
getresp_in_buf (int s, char *buf, int buf_size)
{
	if (recvline (s, buf, buf_size) == -1)
		{
#ifdef	SMTPTRACE	/*  GeK - added to track reset connections  */
		if (Smtptrace)
			tprintf ("smtp recv: failed\n");
#endif	/* SMTPTRACE */
		return (-1);
		}
	
	rip (buf);								 /* Remove cr/lf */
#ifdef	SMTPTRACE
	if (Smtptrace)
		tprintf ("smtp recv: %s\n", buf);	 /* Display to user */
#endif
	return (atoi (buf));
}

/* Wait for, read and display response from server. Return the result code. */
static int
getresp (struct smtpcli *cb,
         int mincode,			/* Keep reading until at least this code comes back */
         char **response)		/* Stores server response if not NULL */
{
	int rval;
	char line[LINELEN];

	usflush (cb->s);
	for (;;)
		{
		/* Get line */
		rval = getresp_in_buf (cb->s, line, LINELEN);
		if (rval < 0)
			break;

#ifdef	LOG_SMTPCLI_ERRORS	/*  GeK - log unexpected responses  */
		if (rval >= 400)
			{
			if (cb->buf[0] != '\0')
				{
				rip (cb->buf);
				log (cb->s, "smtpcli sent: %s", cb->buf);
				}
			log (cb->s, "smtpcli recv: %s", line);
			}
#endif	/* LOG_SMTPCLI_ERRORS */

		if (rval >= 500)
			{									 /* Save permanent error
												  * replies */
			char tmp[LINELEN];

			if (cb->errlog == NULLLIST)
				{
				char	*cp;
				
				/*
				 *	GeK - added resolved mailhost name and IP addr to show
				 *	actual host address and MX dest.  Given resolve_a uses A
				 *	records in cache, it adds little DNS overhead.
				 */
				cp = resolve_a (cb->ipdest, 0);
				/*  GeK - change trailing '.' to ' ' for neatness  */
				if (cp != NULLCHAR && cp[strlen (cp) - 1] == '.')
					cp[strlen (cp) - 1] = ' ';
				sprintf (tmp, "While talking to %s[%s] for %s:",
				         (cp != NULLCHAR ? cp : ""),
				         inet_ntoa (cb->ipdest), cb->destname);
				free (cp);	/*  GeK - free mailhost name  */
				logerr (cb, tmp);
				}

			if (cb->buf[0] != '\0')
				{								 /* Save offending command */
				rip (cb->buf);
				sprintf (tmp, ">>> %s", cb->buf);
				logerr (cb, tmp);
				cb->buf[0] = '\0';
				}

			sprintf (tmp, "<<< %s", line);
			logerr (cb, tmp);					 /* save the error reply */
			}

		/* Messages with dashes are continued */
		if (line[3] != '-' && rval >= mincode)
			{
			/*  GeK - store server response  */
			if (response != NULL)
				*response = strdup (line);
			break;
			}

		}
	return rval;
}

#ifdef	USE_ESMTP
/*
 *	GeK - version of getresp() to handle EHLO/RSET response
 *	(doesn't save errors in cb)
 *	Keeps reading until at least mincode comes back
 */
static int
get_ehlo_resp (struct smtpcli *cb, int mincode)
{
	int     rval;
	char    line [LINELEN];

	usflush (cb->s);
	for (;;)
	{
		/* Get line */
		rval = getresp_in_buf (cb->s, line, LINELEN);
		if (rval < 0)
			break;

#ifdef	LOG_SMTPCLI_ERRORS
		/*  GeK - log unexpected responses (excluding 5xx for EHLO) */
		if (rval >= 400 && rval < 500)
			{
			if (cb->buf[0] != '\0')
				{
				rip (cb->buf);
				log (cb->s, "smtpcli sent: %s", cb->buf);
				}
			log (cb->s, "smtpcli recv: %s", line);
			}
#endif	/* LOG_SMTPCLI_ERRORS */

		if (line[3] != '-' && rval >= mincode)
			break;
	}

	return (rval);
}
#endif	/* USE_ESMTP */

static int
    doverbose (argc, argv, p)
int argc;
char *argv[];
void *p;

	{
	return setbool (&smtpverbose, "SMTP verbose mode", argc, argv);
	}

/*
 *  GeK - Expire delay lock on dir/pre.suf if older than age
 *  Returns:
 *		-2 if delay lock exists but not deleted
 *		-1 on failure to delete existing delay lock
 *		 0 if delay lock deleted successfully
 *		 1 if delay lock doesn't exist
 */
static int
explock (char *dir, char *pre, char *suf, int32 age)
{
	char		fn_buf[80];
	long		difference;
	struct stat	stat_buf;

	sprintf (fn_buf, "%s/%s.%s", dir, pre, suf);
	if (fstat_file (fn_buf, &stat_buf))
		return (1);

	/*  No delay channel, expires immediately  */
	if (Delaytime == 0)
		return (unlink (fn_buf));

	difference = time (NULL) - stat_buf.st_ctime;
	if (difference >= age)
		return (unlink (fn_buf));

	/*  Exists but not deleted  */
	return (-2);
}

/*
 *  GeK - helper function to replace stat() - Borland's stat in certain
 *  circumstances calls realloc() which is unimplemented in alloc.c
 *  BC 3.1 stat() also leaks mem (fix on ftp.borland.com)
 *  fstat() seems OK, it's also used in domain.c
 *  Returns 0 on success, -1 on failure.
 */
static int
fstat_file (const char *filepath, struct stat *statp)
{
	int	fhandle;
	int result;

	fhandle = open (filepath, O_RDONLY);
	if (fhandle == -1)
		return (-1);

	result = fstat (fhandle, statp);
	close (fhandle);

	return (result);
}
